package moonlander

import moonlander.models.Ground
import moonlander.models.MoonLander
import org.joml.Vector2f
import java.util.*
import kotlin.math.absoluteValue


/**
 * Main class containing all the game logic.
 * Every frame (=tick), it triggers physics simulation and checks if the game is won/lost/over.
 */
class Controller {
	val moonLander = MoonLander()
	val ground = Ground(0f, 9.0)
	var totalTime = -3.0
	var gameStarted = false
	var gameIsRunning = false
	var gameIsOver = false
	var gameIsLost = false
	var gameLostMessage: String = ""

	init {
		reset()
	}

	fun reset() {
		if (gameIsOver && !gameIsLost) {
			// also reset the ground
			ground.randomize()
		}

		totalTime = -3.0
		gameStarted = false
		gameIsRunning = false
		gameIsOver = false
		gameIsLost = false
		gameLostMessage = ""
		stuckCounter = 0.0

		lastTick = 0L
		landerEngineAfterStart = false
		landerSideEngineAfterStart = 0

		moonLander.reset()
		moonLander.randomizeStartPosition()
		moonLander.addInitialOffset(-totalTime.toFloat())
		moonLander.engineFiring = true
		moonLander.useFuel = false
	}


	fun landerHeight() = ground.heightOverGround(moonLander.position)


	fun physics(dt: Float) {
		if (gameIsRunning) {
			val collisions = ground.getCollisions(moonLander)
			moonLander.reportCollisions(collisions)
			if (collisions.size > 0) {
				/*println("${collisions.size}:")
				for (c in collisions)
					println("${c.penetration}  ${c.normal.x} ${c.normal.y}")*/
			}

			collisions.forEach { c -> c.applyImpulse() }
			moonLander.physics(dt)
			totalTime += dt
			collisions.forEach { c -> c.positionalCorrection() }
		}
		if (!gameIsRunning || moonLander.hp <= 0) {
			fireStop()
			sideFireStop()
		}
		if (moonLander.hp <= 0) {
			gameIsLost = true
			gameLostMessage = "lander crashed into the surface"
		}
	}


	private val lastLanderPositions = LinkedList<Vector2f>()
	private val lastDts = LinkedList<Double>()
	private var lastLanderAngle = 0f
	private var endgameCounter = 0.0

	/**
	 * Game is over iff the lander didn't move for more than 2 seconds while all engines have been off.
	 */
	fun checkForGameEnd(dt: Double) {
		if (!gameIsRunning)
			return

		// set endgameCounter to zero if this can't be the end,
		// otherwise increment by dt
		if (moonLander.engineFiring || moonLander.sideEngineFiring != 0 || lastLanderPositions.isEmpty()) {
			endgameCounter = 0.0
		} else {
			val speed = moonLander.position.sub(lastLanderPositions.first, Vector2f()).length() / dt + lastDts.sum()
			val angleSpeed = (lastLanderAngle - moonLander.orientationAngle()).absoluteValue / dt
			if (speed <= 0.8 && angleSpeed <= 1.5) {
				endgameCounter += dt
			} else {
				endgameCounter = 0.0
				println("speed = $speed     angleSpeed = $angleSpeed")
			}
		}

		if (lastLanderPositions.isNotEmpty())
			lastDts.add(dt)
		lastLanderPositions.add(Vector2f(moonLander.position))
		if (lastLanderPositions.size > 15) {
			lastLanderPositions.removeFirst()
			lastDts.removeFirst()
		}
		lastLanderAngle = moonLander.orientationAngle()

		if (endgameCounter > 2.0) {
			// Game is over!
			gameIsRunning = false
			gameIsOver = true
			// check final angle of lander
			if (!gameIsLost && moonLander.orientationAngle().absoluteValue > 45) {
				gameIsLost = true
				gameLostMessage = "lander not standing upright"
			}
		}
	}


	private var stuckCounter = 0.0

	fun checkForStuck(dt: Double) {
		if (!gameIsRunning)
			return

		if ((moonLander.fuel <= 0 || gameIsLost)
			&& moonLander.position.y - moonLander.maxRadius - 2f <= ground.level + ground.maxDiffHeight
			&& moonLander.speed.length() < 3f && moonLander.angularSpeed <= 0.5f
		) {
			stuckCounter += dt
			println("stuckCounter += $dt  =>  $stuckCounter")
		}

		if (stuckCounter > 22.0) {
			gameIsRunning = false
			gameIsOver = true
			if (!gameIsLost) {
				gameIsLost = true
				gameLostMessage = "lander got stuck"
			}
		}
	}


	private var lastTick = 0L

	fun tick() {
		if (lastTick == 0L) {
			lastTick = System.nanoTime()
		} else {
			val t = System.nanoTime()
			val dt = Math.min(0.1, (t - lastTick) / 1000000000.0)
			lastTick = t

			if (totalTime < 0 && totalTime + dt >= 0) {
				// Game starts!
				moonLander.useFuel = true
				moonLander.engineFiring = landerEngineAfterStart
				moonLander.sideEngineFiring = landerSideEngineAfterStart
			}

			physics(dt.toFloat())
			checkForGameEnd(dt)
			checkForStuck(dt)
		}
	}


	private var landerEngineAfterStart = false // main engine has been fired while game was not started yet (e.g. user holds space pressed)
	private var landerSideEngineAfterStart = 0 // side engine has been fired while game was not started yet

	fun fireStart() {
		if (totalTime >= 0) {
			if (gameIsRunning && moonLander.hp > 0)
				moonLander.engineFiring = true
		} else {
			landerEngineAfterStart = true
		}
	}

	fun fireStop() {
		if (totalTime >= 0)
			moonLander.engineFiring = false
		else
			landerEngineAfterStart = false
	}

	fun sideFireStart(isLeft: Boolean) {
		if (totalTime >= 0) {
			if (gameIsRunning && moonLander.hp > 0)
				moonLander.sideEngineFiring = if (isLeft) 1 else -1
		} else {
			landerSideEngineAfterStart = if (isLeft) 1 else -1
		}
	}

	fun sideFireStop() {
		if (totalTime >= 0)
			moonLander.sideEngineFiring = 0
		else
			landerSideEngineAfterStart = 0
	}
}