package moonlander

import moonlander.views.BackgroundView
import moonlander.views.GroundView
import moonlander.views.MoonLanderView
import moonlander.views.hud.HUDView
import org.joml.Matrix4f
import org.joml.Vector2f
import org.lwjgl.glfw.GLFW
import org.lwjgl.opengl.GL
import org.lwjgl.opengl.GL30.*
import java.awt.SystemColor.window


class MoonLandingApp(val controller: Controller) : AbstractApp() {

	// HUD coordinate system: proportional to window size, (0,0) is top left
	var hudMatrix = Matrix4f()
	var hudWidth = 0f
	var hudHeight = 0f

	// View space: proportional to window size, boundaries determined later (by lander position). Transforms world space -> view space.
	var viewMatrix = Matrix4f()
	val hTop = 100f
	val hBottom = -6f
	var center = 0f
	var hf = (hTop - hBottom)
	var wf = 0f


	private var onReset: (() -> Unit)? = null

	override fun onKeyPress(window: Long, key: Int, scancode: Int, action: Int, mods: Int) {
		if (!controller.gameStarted) {
			// Start the game
			if (key == GLFW.GLFW_KEY_SPACE && action == GLFW.GLFW_PRESS) {
				controller.gameStarted = true
				controller.gameIsRunning = true
			}
		} else if (controller.gameIsOver || controller.gameIsLost) {
			// Game is over, you might restart
			if (key == GLFW.GLFW_KEY_ENTER && action == GLFW.GLFW_PRESS) {
				controller.reset()
				onReset?.invoke()
				center = 0f
			}
		} else {
			// Key controls
			when (key) {
				// main engine
				GLFW.GLFW_KEY_SPACE, GLFW.GLFW_KEY_W -> {
					when (action) {
						GLFW.GLFW_PRESS -> controller.fireStart()
						GLFW.GLFW_RELEASE -> controller.fireStop()
					}
				}
				// side engine left
				GLFW.GLFW_KEY_A, GLFW.GLFW_KEY_LEFT -> {
					when (action) {
						GLFW.GLFW_PRESS -> controller.sideFireStart(true)
						GLFW.GLFW_RELEASE -> controller.sideFireStop()
					}
				}
				// side engine right
				GLFW.GLFW_KEY_D, GLFW.GLFW_KEY_RIGHT -> {
					when (action) {
						GLFW.GLFW_PRESS -> controller.sideFireStart(false)
						GLFW.GLFW_RELEASE -> controller.sideFireStop()
					}
				}
			}
		}
	}

	override fun updateWindowSize(w: Int, h: Int) {
		// Update viewport matrix. Visible range [top/bottom] is fixed, viewport width is adjusted.
		// Currently, viewport is always centered
		wf = hf / h * w

		// HUD - (0,0) is top-left
		hudHeight = 3000f
		hudWidth = hudHeight / h * w
		hudMatrix.identity().ortho2D(0f, hudWidth, -hudHeight, 0f)
	}

	fun setViewMatrix(position: Vector2f) {
		if (center - wf / 2f > position.x - 20)
			center = position.x - 20 + wf / 2
		if (center + wf / 2f < position.x + 20)
			center = position.x + 20 - wf / 2
		viewMatrix.identity().ortho2D(center - wf / 2f, center + wf / 2f, hBottom, hTop)
	}

	override fun loop() {
		GL.createCapabilities()

		val lander = MoonLanderView(controller, controller.moonLander)
		val ground = GroundView(controller.ground)
		val background = BackgroundView()
		background.preloadStart()
		val hud = HUDView(controller)

		onReset = {
			ground.reset()
		}

		// Set the clear color
		glClearColor(0.0f, 0.05f, 0.3f, 0.0f)

		// Run the rendering loop until the user has attempted to close the window or has pressed the ESCAPE key.
		while (!GLFW.glfwWindowShouldClose(window)) {
			controller.tick()
			setViewMatrix(controller.moonLander.position)

			// clear the framebuffer
			glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT)

			// Draw things (background first)
			background.render(viewMatrix)
			ground.render(viewMatrix)
			lander.render(viewMatrix)
			hud.render(hudMatrix, hudWidth, hudHeight)

			// Show frame
			GLFW.glfwSwapBuffers(window)
			GLFW.glfwPollEvents()// The key callback above will only be invoked during this call.
		}
	}
}
