package moonlander.models

import org.joml.Vector2f
import java.util.*
import kotlin.math.abs
import kotlin.math.max

// Everything related to collision detection is here.
// Check this tutorial for more info:
// https://gamedevelopment.tutsplus.com/tutorials/how-to-create-a-custom-2d-physics-engine-oriented-rigid-bodies--gamedev-8032


/**
 * A (convex) polygon shape (e.g. a "hitbox") for an object.
 * Points must be in counter-clockwise order!
 * Polygon is not closed, except if last point = first point.
 * All coordinates are world-space (in contrast to the tutorial)
 */
open class CollisionPolygon(val obj: PhysicalObject?) {
	val points = ArrayList<Vector2f>()
	var normals: List<Vector2f> = ArrayList() // |normals| = |points|-1

	constructor(obj: PhysicalObject?, initialPoints: Iterable<Vector2f>) : this(obj) {
		initialPoints.forEach { points.add(it) }
		update()
	}

	fun update() {
		normals = (0 until points.size - 1).map { i ->
			Vector2f(points[i + 1].y - points[i].y, points[i].x - points[i + 1].x).normalize()
		}.toList()
	}

	/**
	 * Left boundary of this polygon
	 */
	fun min() = points.map { it.x }.min() ?: throw RuntimeException("Points is empty")

	/**
	 * Right boundary of this polygon
	 */
	fun max() = points.map { it.x }.max() ?: throw RuntimeException("Points is empty")

	/**
	 * Central point of this polygon. Simply: average of all points.
	 */
	fun centerOfMass(): Vector2f =
		obj?.position ?: Vector2f(points.map { it.x }.average().toFloat(), points.map { it.y }.average().toFloat())

	/**
	 * 1/m (or 0 if the object is not able to move)
	 */
	fun inverseMass(): Float = if (obj != null) 1 / obj.mass else 0f

	/**
	 * 1/i (or 0 if the object is not able to move)
	 */
	fun inverseInertia(): Float = if (obj != null) 1 / obj.inertia else 0f

	@Suppress("unused")
	fun toSvg(color: String = "blue"): String {
		// SVG polygon (for debugging purposes)
		return "<polygon points=\"" +
				points.joinToString(" ") { "${it.x},${-it.y}" } +
				"\" style=\"fill:$color;stroke:black;stroke-width:0.1\" opacity=\"0.5\" />\n"
	}

	/**
	 * Find the point which lies most far away (from origin) in the given direction
	 */
	fun findSupportPoint(dir: Vector2f): Vector2f {
		return points.maxBy { p -> p.dot(dir) } ?: throw RuntimeException("Points is empty")
	}

	open fun findAxisLeastPenetration(B: CollisionPolygon): Pair<Float, Int> {
		var dMax = Float.NEGATIVE_INFINITY
		var index = -1
		for (i in 0 until normals.size) {
			val negativeNormal = Vector2f(normals[i]).negate()
			val support = B.findSupportPoint(negativeNormal)
			val d = normals[i].dot(Vector2f(support).sub(points[i]))
			if (d > dMax) {
				dMax = d
				index = i
			}
		}

		return Pair(dMax, index)
	}

	fun getFace(index: Int) = Face(points[index], points[index + 1])

	/**
	 * Called on RefPoly
	 */
	fun findIncidentFace(incPoly: CollisionPolygon, refFaceIndex: Int): Face {
		val refNormal = normals[refFaceIndex]
		// Find most anti-normal face on incident polygon
		val incidentFaceIndex = (0 until incPoly.normals.size).minBy { i -> refNormal.dot(incPoly.normals[i]) }
		return incPoly.getFace(incidentFaceIndex ?: throw RuntimeException("No faces found!"))
	}

}


/**
 * Special collision polygon of a (piece of) the ground.
 * Speciality: Objects might only collide with the surface of the ground piece (not with anything under the ground)
 */
class GroundCollisionPolygon : CollisionPolygon(null) {
	override fun findAxisLeastPenetration(B: CollisionPolygon): Pair<Float, Int> {
		var dMax = Float.NEGATIVE_INFINITY
		var index = -1
		var dMax2 = Float.NEGATIVE_INFINITY
		var index2 = -1

		for (i in 0 until normals.size) {
			val negativeNormal = Vector2f(normals[i]).negate()
			val support = B.findSupportPoint(negativeNormal)
			val d = normals[i].dot(Vector2f(support).sub(points[i]))
			if (d > dMax) {
				dMax = d
				index = i
			}
			if (i < normals.size - 3 && d > dMax2) {
				dMax2 = d
				index2 = i
			}
		}

		if (dMax < 0f) {
			return Pair(dMax2, index2)
		}

		return Pair(dMax, index)
	}
}


/**
 * A summary of a collision between two polygons A and B.
 * Heavily inspired by the tutorial's example code
 * (not to say: Mr. Guttenberg would be proud of this)
 */
class Manifold(val A: CollisionPolygon, val B: CollisionPolygon) {
	val contacts = ArrayList<Vector2f>()
	var normal = Vector2f()
	var penetration = 0f

	val e = 0.05f // 0.2f
	val staticFriction = 0.7f
	val dynamicFriction = 0.5f

	fun applyImpulse() {
		val sf = staticFriction
		val df = dynamicFriction

		for (c in contacts) {
			val ra = c.sub(A.centerOfMass(), Vector2f())
			val rb = c.sub(B.centerOfMass(), Vector2f())
			// Relative velocity
			val objA = A.obj
			val objB = B.obj
			var vb = if (objB != null) objB.speed.add(cross(objB.angularSpeed, rb), Vector2f()) else Vector2f()
			var va = if (objA != null) objA.speed.add(cross(objA.angularSpeed, ra), Vector2f()) else Vector2f()
			var rv = vb.sub(va)
			// Relative velocity along the normal
			val contactVel = rv.dot(normal)
			if (contactVel > 0)
				return

			val raCrossN = cross(ra, normal)
			val rbCrossN = cross(rb, normal)
			val invMassSum = A.inverseMass() + B.inverseMass() +
					raCrossN * raCrossN * A.inverseInertia() + rbCrossN * rbCrossN * B.inverseInertia()

			// Calculate impulse scalar
			val j = -(1f + e) * contactVel / invMassSum / contacts.size
			val impulse = Vector2f(normal).mul(j)
			A.obj?.applyImpulse(impulse.negate(Vector2f()), ra)
			B.obj?.applyImpulse(impulse, rb)

			// Friction impulse
			vb = if (objB != null) objB.speed.add(cross(objB.angularSpeed, rb), Vector2f()) else Vector2f()
			va = if (objA != null) objA.speed.add(cross(objA.angularSpeed, ra), Vector2f()) else Vector2f()
			rv = vb.sub(va)

			val t = rv.sub(normal.mul(rv.dot(normal), Vector2f()), Vector2f()).normalize()

			// j tangent magnitude
			val jt = -rv.dot(t) / invMassSum / contacts.size

			// Don't apply tiny friction impulses
			if (abs(jt) < 0.0001f)
				return

			// Coulumb's law
			val tangentImpulse = if (abs(jt) < j * sf) t.mul(jt) else t.mul(-j * df)

			// Apply friction impulse
			A.obj?.applyImpulse(tangentImpulse.negate(Vector2f()), ra)
			B.obj?.applyImpulse(tangentImpulse, rb)
		}
	}

	fun positionalCorrection() {
		val correction = Vector2f(normal)
		correction.mul(max(penetration - 0.05f, 0f) / (A.inverseMass() + B.inverseMass()))
		correction.mul(0.4f)
		A.obj?.position?.add(correction.mul(A.inverseMass(), Vector2f()))
		B.obj?.position?.add(correction.mul(B.inverseMass(), Vector2f()))
	}
}

data class Face(var a: Vector2f, var b: Vector2f)

fun biasGreaterThan(a: Float, b: Float) = a >= b * 0.95f + a * 0.01f

fun cross(a: Float, v: Vector2f) = Vector2f(-a * v.y, a * v.x)

fun cross(a: Vector2f, b: Vector2f): Float = a.x * b.y - a.y * b.x


fun clip(n: Vector2f, c: Float, face: Face): Int {
	var sp = 0
	val out = arrayOf(face.a, face.b)

	// Retrieve distances from each endpoint to the line
	// d = ax + by - c
	val d1 = n.dot(face.a) - c
	val d2 = n.dot(face.b) - c

	// If negative (behind plane) clip
	if (d1 <= 0.0f) out[sp++] = face.a
	if (d2 <= 0.0f) out[sp++] = face.b

	// If the points are on different sides of the plane
	if (d1 * d2 < 0.0f) {
		// Push intersection point
		val alpha = d1 / (d1 - d2)
		out[sp] = Vector2f(face.b).sub(face.a).mul(alpha).add(face.a)
		++sp
	}

	face.a = out[0]
	face.b = out[1]

	assert(sp != 3)

	return sp
}


/**
 * Check if two polygons collide.
 * @return a manifold summarizing the collision, or null if no collision happens.
 */
fun checkForCollisions(A: CollisionPolygon, B: CollisionPolygon): Manifold? {
	val (penetrationA, faceA) = A.findAxisLeastPenetration(B)
	if (penetrationA >= 0f)
		return null
	val (penetrationB, faceB) = B.findAxisLeastPenetration(A)
	if (penetrationB >= 0f)
		return null

	var refPoly = A
	var incPoly = B
	var referenceFaceIndex = faceA
	var flip = false
	if (!biasGreaterThan(penetrationA, penetrationB)) { // BiasGreaterThan???
		refPoly = B
		incPoly = A
		referenceFaceIndex = faceB
		flip = true
	}
	val incidentFace = refPoly.findIncidentFace(incPoly, referenceFaceIndex)

	val v1 = refPoly.points[referenceFaceIndex]
	val v2 = refPoly.points[referenceFaceIndex + 1]
	val sidePlaneNormal = Vector2f(v2).sub(v1).normalize()
	val refFaceNormal = Vector2f(sidePlaneNormal.y, -sidePlaneNormal.x)

	// ax + by = c
	// c is distance from origin
	val refC = refFaceNormal.dot(v1)
	val negSide = -sidePlaneNormal.dot(v1)
	val posSide = sidePlaneNormal.dot(v2)

	// Clip incident face to reference face side planes
	if (clip(sidePlaneNormal.negate(Vector2f()), negSide, incidentFace) < 2) {
		// println("c1")
		return null // Due to floating point error, possible to not have required points
	}
	if (clip(sidePlaneNormal, posSide, incidentFace) < 2) {
		// println("c2")
		return null // Due to floating point error, possible to not have required points
	}

	val manifold = Manifold(A, B)
	manifold.normal = Vector2f(refFaceNormal)
	if (flip) manifold.normal.negate()

	// Keep points behind reference face
	var separation = refFaceNormal.dot(incidentFace.a) - refC
	if (separation <= 0f) {
		manifold.contacts.add(incidentFace.a)
		manifold.penetration += -separation
	}
	separation = refFaceNormal.dot(incidentFace.b) - refC
	if (separation <= 0f) {
		manifold.contacts.add(incidentFace.b)
		manifold.penetration += -separation
	}
	if (manifold.contacts.size > 0)
		manifold.penetration /= manifold.contacts.size
	return manifold
}