package moonlander.models

import org.joml.Vector2f
import OpenSimplexNoise
import moonlander.PerlinNoise
import kotlin.math.floor
import kotlin.math.roundToInt
import kotlin.random.Random


/**
 * This class represents the ground of the moon.
 * The height is stored as list of heights (left to right). These height values are grouped into buckets (which will be generated at once.
 * Each bucket is "bucketWidth" (=80m) wide, and consists of "bucketSize" =64 sampled points.
 * Bucket #0 starts at x=0 and goes up to x=80, bucket #1 starts at x=80 and goes up to x=160...
 *
 */
class Ground(val level: Float = 0f, private val stretch: Double = 1.0) {
	/**
	 * Amplitude of terrain height. All height values in buckets will be in [level-maxDiffHeight : level+maxDiffHeight]
	 */
	val maxDiffHeight = 6f

	val bucketWidth = 80f

	val bucketSize = 64

	// Ground position (x,y) where x is index:
	// y = buckets[x//64][x%64]
	private var noiseGenerator = OpenSimplexNoise(Random.Default.nextLong()) // "5" is my special test-ground
	private val buckets: HashMap<Int, Array<Float>> = HashMap() // cache generated buckets

	/**
	 * Retrieve the bucket with the given index
	 * = an array containing height values for a piece of the ground
	 */
	fun getGroundBucket(index: Int): Array<Float> {
		return buckets.getOrPut(index) {
			Array(bucketSize + 1) { i ->
				var result = 0.75 * noiseGenerator.eval((index * bucketSize + i).toDouble() / stretch, 0.0)
				result += 0.25 * noiseGenerator.eval((index * bucketSize + i).toDouble() * 3.0 / stretch, 1000.0)
				result.toFloat() * maxDiffHeight
			}
		}
	}

	/**
	 * Use this noise generator to generate the texture of the terrain
	 */
	val perlinNoiseGenerator = PerlinNoise()

	init {
		perlinNoiseGenerator.octaves.add(Pair(.5f, 1 / 4f))
		perlinNoiseGenerator.octaves.add(Pair(.25f, 2 / 4f))
		perlinNoiseGenerator.octaves.add(Pair(.12f, 4 / 4f))
		perlinNoiseGenerator.octaves.add(Pair(.06f, 8 / 4f))
		perlinNoiseGenerator.octaves.add(Pair(.03f, 16 / 4f))
	}


	/**
	 * Height at position X in a given bucket (where X is [0 .. bucketSize]
	 */
	fun heightAtIndex(bucketNumber: Int, fi: Float): Float {
		val bucket = getGroundBucket(bucketNumber)
		val bucketIndex = floor(fi).toInt()
		val bucketRemainder = fi - bucketIndex
		return (1 - bucketRemainder) * bucket[bucketIndex] + bucketRemainder * bucket[bucketIndex + 1]
	}


	/**
	 * Height at position X (where X is in ground index space, not world coordinates)
	 */
	fun heightOverGround(position: Vector2f): Float {
		val x1 = floor(position.x).toInt()
		val i = ((x1 % bucketSize) + bucketSize) % bucketSize
		val bucket = getGroundBucket(x1 / bucketSize)
		val h = bucket[i] * (x1 + 1 - position.x) + bucket[i + 1] * (position.x - x1)
		return position.y - (level + h)
	}


	/**
	 * Check if a physical object collides with the ground.
	 * Returns a list of all collisions. If the result is an empty list, no collision happened.
	 */
	fun getCollisions(obj: PhysicalObject): List<Manifold> {
		// pre-check if object is deep enough
		if (obj.position.y - (level + maxDiffHeight) > obj.maxRadius)
			return listOf()

		// transform polygon points to world space
		val objPoly = obj.getCollisionPoly()
		val groundPolys = getCollisionPolys2(objPoly.min(), objPoly.max())

		// Get list with all collisions
		return groundPolys.mapNotNull { groundPoly -> checkForCollisions(groundPoly, objPoly) }
	}

	/**
	 * Simple method to generate a collision polygon
	 */
	@Suppress("unused")
	private fun getCollisionPolys(min: Float, max: Float): List<CollisionPolygon> {
		val minIndex = floor(min / bucketWidth * bucketSize).roundToInt()
		val maxIndex = floor(max / bucketWidth * bucketSize).roundToInt()

		return (minIndex..maxIndex).map { i ->
			val groundPoly = GroundCollisionPolygon()
			val innerIndex = ((i % bucketSize) + bucketSize) % bucketSize
			val b = getGroundBucket((i - innerIndex) / bucketSize)
			val y1 = level + b[innerIndex]
			val y2 = level + b[innerIndex + 1]
			val x1 = i * bucketWidth / bucketSize
			val x2 = (i + 1) * bucketWidth / bucketSize
			// Close the polygon
			groundPoly.points.add(Vector2f(x2, y2))
			groundPoly.points.add(Vector2f(x1, y1))
			groundPoly.points.add(Vector2f(x1, level - 2 * maxDiffHeight))
			groundPoly.points.add(Vector2f(x2, level - 2 * maxDiffHeight))
			groundPoly.points.add(Vector2f(groundPoly.points[0]))
			// Done with ground points
			groundPoly.update()
			groundPoly
		}.toList()
	}

	/**
	 * Generate an advanced collision polygon for a part of the ground.
	 * This method tries to combine as many pieces into one polygon as possible, while still creating convex polys.
	 */
	private fun getCollisionPolys2(min: Float, max: Float): List<CollisionPolygon> {
		val minIndex = floor(min / bucketWidth * bucketSize).roundToInt()
		val maxIndex = floor(max / bucketWidth * bucketSize).roundToInt()

		val polys = ArrayList<CollisionPolygon>()
		var poly: GroundCollisionPolygon? = null
		var lastM = 0f
		var startX = 0f

		for (i in minIndex..maxIndex) {
			val innerIndex = ((i % bucketSize) + bucketSize) % bucketSize
			val b = getGroundBucket((i - innerIndex) / bucketSize)
			val y1 = level + b[innerIndex]
			val y2 = level + b[innerIndex + 1]
			val x1 = i * bucketWidth / bucketSize
			val x2 = (i + 1) * bucketWidth / bucketSize
			val m = y2 - y1

			if (poly == null) {
				// Open poly
				poly = GroundCollisionPolygon()
				// add "previous-interpolated" point
				startX = x1 - (x2 - x1)
				poly.points.add(Vector2f(startX, y1 - m))
				lastM = m
			}

			// Add current point to poly
			poly.points.add(0, Vector2f(x1, y1))

			if (m > lastM) {
				// Close poly, including interpolated point
				poly.points.add(0, Vector2f(x2, y1 + lastM))
				poly.points.add(Vector2f(startX, level - 2 * maxDiffHeight))
				poly.points.add(Vector2f(x2, level - 2 * maxDiffHeight))
				poly.points.add(poly.points[0])
				poly.update()
				polys.add(poly)
				poly = null
			}

			if (poly == null) {
				// Open poly
				poly = GroundCollisionPolygon()
				// add "previous-interpolated" point
				startX = x1 - (x2 - x1)
				poly.points.add(Vector2f(startX, y1 - m))
				poly.points.add(0, Vector2f(x1, y1))
			}

			lastM = m
		}

		if (poly != null) {
			// Close poly
			poly.points.add(Vector2f(startX, level - 2 * maxDiffHeight))
			poly.points.add(Vector2f(maxIndex * bucketWidth / bucketSize, level - 2 * maxDiffHeight))
			poly.points.add(poly.points[0])
			poly.update()
			polys.add(poly)
		}

		return polys
	}


	fun randomize() {
		noiseGenerator = OpenSimplexNoise(Random.Default.nextLong())
		buckets.clear()
	}

}