package moonlander.models

import org.joml.Matrix3x2f
import org.joml.Vector2f
import kotlin.math.PI
import kotlin.math.absoluteValue
import kotlin.math.floor
import kotlin.math.max
import kotlin.random.Random

/**
 * The moon lander.
 * Contains advanced physics calculation (engines etc), fuel control, collision control and damage calculation.
 */
class MoonLander : PhysicalObject() {
	val engineAcceleration = 3.5f // m/s² in flight direction
	val engineFuelConsumption = 1f // liter / second burn
	val sideEngineAcceleration = 1f // rad/s²
	val sideEngineFuelConsumption = 0.1f // liter / second burn
	val maxFuel = 20f // liter

	var fuel = maxFuel
	var useFuel = true // engine does not use fuel before the game starts

	var engineFiring = false
	var sideEngineFiring = 0 // 0 = off, 1 = left, -1 = right

	init {
		reset()

		// border points - for now in texture image format. 0 and 2 are the landing barks, 1 is the main engine.
		borderPoints.add(Vector2f(20f, 801f))
		borderPoints.add(Vector2f(775f, 815f))
		borderPoints.add(Vector2f(775f, 700f))
		borderPoints.add(Vector2f(554f, 323f))
		borderPoints.add(Vector2f(464f, 95f))
		borderPoints.add(Vector2f(234f, 312f))
		borderPoints.add(Vector2f(20f, 700f))
		borderPoints.add(Vector2f(20f, 801f))
		// convert to real coordinates
		val texWidth = 800f
		val texHeight = 888f
		borderPoints.onEach { v ->
			v.x = (v.x - texWidth / 2f) * size * 2f / texHeight
			v.y = (texHeight / 2f - v.y) * size * 2f / texHeight
		}
		updatedBorderPoints()
	}

	/**
	 * Move the lander to its original position
	 */
	override fun reset() {
		super.reset()

		fuel = maxFuel
		useFuel = true
		engineFiring = false
		sideEngineFiring = 0

		position.set(0f, 90f)
		speed.set(0f, -12f)

		maxCollisionSpeedUpperLander = 0f
		maxCollisionSpeedStands = 0f
		hp = 100
	}

	override fun physics(dt: Float) {
		// Gravity
		acceleration.add(0f, lunarGravity)

		// Main Engine
		if (engineFiring && fuel > 0) {
			// accelerate in the direction of the orientation
			val acc = Vector2f(0f, engineAcceleration)
			acc.mulDirection(orientation)
			acceleration.add(acc)
			if (useFuel)
				fuel -= engineFuelConsumption * dt
		}

		// Side engine
		if (sideEngineFiring != 0) {
			// rotational acceleration
			torque += sideEngineAcceleration * sideEngineFiring
			fuel -= sideEngineFuelConsumption * dt
		}

		// Fuel
		if (fuel <= 0f) {
			engineFiring = false
			sideEngineFiring = 0
			fuel = 0f
		}

		super.physics(dt)
	}


	var maxCollisionSpeedStands = 0f
	var maxCollisionSpeedUpperLander = 0f

	val standCollisionSpeedLimit = 2f
	val landerCollisionSpeedLimit = 0.10f

	/**
	 * Given a list of collision, this method checks if any collisions might damage the lander.
	 * Lander has two zones: stand (bottom part) and payload (upper part). Stand is designed to withstand smaller impacts, payload is quite fragile.
	 * Manifolds are always (ground, lander).
	 */
	fun reportCollisions(collisions: List<Manifold>) {
		var standCollisionSpeed = 0f
		var landerCollisionSpeed = 0f

		// Find hardest collision from that list
		for (manifold in collisions) {
			// "averaged" contact point
			val contactPoint = manifold.contacts.fold(Vector2f()) { a, b -> a.add(b) }.mul(1f / manifold.contacts.size)
			val collisionSpeedVector = cross(angularSpeed, contactPoint.sub(position))
			// if no exact collision point is known, we don't have any rotational speed
			if (collisionSpeedVector.length().isNaN())
				collisionSpeedVector.zero()
			// collision speed = rotation speed of point + general speed
			collisionSpeedVector.add(speed)

			// check speed to see if anything got broken due to the shock. We use speed in direction of the collision normal
			val collisionSpeed = manifold.normal.dot(collisionSpeedVector).absoluteValue

			// check which side of the lander got in contact (bottom = landing stands, everything above = not designed to touch ground)
			val maxLanderY = manifold.contacts.map { point ->
				// Convert collision point to model space and take y coordinate
				Vector2f(point).sub(position).mulDirection(orientation.invert(Matrix3x2f())).y
			}.max()?.div(size) ?: continue

			if (maxLanderY <= -0.3f) {
				standCollisionSpeed = max(standCollisionSpeed, collisionSpeed)
			} else {
				landerCollisionSpeed = max(landerCollisionSpeed, collisionSpeed)
			}
		}

		// stand collision over tolerable limit?
		if (standCollisionSpeed > maxCollisionSpeedStands || standCollisionSpeed > standCollisionSpeedLimit) {
			maxCollisionSpeedStands = max(maxCollisionSpeedStands, standCollisionSpeed)
			println("HIT! Stands with $standCollisionSpeed m/s!")
			hit(standCollisionSpeed, standCollisionSpeedLimit, 2 * standCollisionSpeedLimit)
		}
		// payload collision over tolerable limit?
		if (landerCollisionSpeed > maxCollisionSpeedUpperLander || landerCollisionSpeed > landerCollisionSpeedLimit) {
			maxCollisionSpeedUpperLander = max(maxCollisionSpeedUpperLander, landerCollisionSpeed)
			println("HIT! Lander with $landerCollisionSpeed m/s!")
			hit(landerCollisionSpeed, landerCollisionSpeedLimit, 3 * landerCollisionSpeedLimit)
		}
	}


	var hp = 100 // hitpoints

	/**
	 * Calculate the damage of a ground collision.
	 * Called every time an interesting collision happens. That means: Scratching the ground will quickly destroy everything!
	 */
	private fun hit(speed: Float, limit: Float, destructionSpeed: Float) {
		if (speed > limit) {
			val lostHp = (speed - limit) / (destructionSpeed - limit) * 100f
			hp = max(0, hp - floor(lostHp).toInt())
			println("hp = $hp")
		}
	}


	/**
	 * Random direction (+- 45°) when the game starts. Corrects orientation and speed.
	 */
	fun randomizeStartPosition() {
		orientation.rotate((Random.Default.nextFloat() * 2f - 1) * PI.toFloat() / 4f)
		speed.mulDirection(orientation)
	}

	/**
	 * Move and accelerate the lander so that it will reach its original position after dt seconds (with burning main engine).
	 * Calculates the movement over the initial countdown.
	 */
	fun addInitialOffset(dt: Float) {
		val acc = Vector2f(0f, engineAcceleration)
		acc.mulDirection(orientation)
		acc.y += lunarGravity
		val deltaV = acc.mul(dt, Vector2f())
		val deltaS = Vector2f(speed).add(speed).sub(deltaV).mul(dt / 2f)

		position.sub(deltaS)
		speed.sub(deltaV)
	}
}