package moonlander.models

import org.joml.Matrix3x2f
import org.joml.Vector2f
import kotlin.math.PI
import kotlin.math.acos
import kotlin.math.sqrt

/**
 * An abstract object containing basic physics (position, speed, gravity etc).
 */
open class PhysicalObject {
	val lunarGravity = -1.62f // m/s² in y-direction

	val size = 5f // in m - RADIUS!
	val mass = 10334f // in kg
	val inertia: Float = mass * size * size * 2f

	// Positional components
	val position = Vector2f() // in m (x,y), (initial) position of the central point (0,0)
	val speed = Vector2f() // in m/s
	val acceleration = Vector2f() // in m/s². all physical elements add acceleration, which will be zero'd at the end of a physics() cycle

	// Angular components
	var orientation = Matrix3x2f()  // radians
	var angularSpeed = 0f // in rad / second, counter-clockwise
	var torque = 0f // acceleration, not force! (e.g. inertia not considered)


	/**
	 * Set initial values (of position etc.)
	 */
	open fun reset() {
		orientation.identity()
		angularSpeed = 0f
		torque = 0f
	}


	/**
	 * Override this method, modify acceleration and torque (by the applying forces), then call inherited method
	 */
	open fun physics(dt: Float) {
		// Apply acceleration
		speed.add(acceleration.x * dt, acceleration.y * dt)
		acceleration.zero()

		// Apply speed
		// pos += speed * dt
		position.add(speed.x * dt, speed.y * dt)

		// Apply Rotation
		angularSpeed += torque * dt
		orientation.rotate(angularSpeed * dt)
		torque = 0f
	}

	/**
	 * The current orientation in degrees
	 */
	fun orientationAngle(): Float {
		val vec = Vector2f(0f, 1f)
		vec.mulDirection(orientation)
		return acos(vec.y) * 180f / PI.toFloat()
	}

	fun applyImpulse(impulse: Vector2f, contactVector: Vector2f) {
		speed.add(Vector2f(impulse).mul(1f / mass))
		angularSpeed += cross(contactVector, impulse) / inertia
	}


	/**
	 * Border points for collision detection.
	 * Coordinate unit: m, where (0,0) is the center of the object. Consider size when calculating
	 * Last point = first point
	 * Points must be given counter-clockwise, the shape must be convex.
	 */
	val borderPoints = ArrayList<Vector2f>()
	var maxRadiusSquare = 0f
	var maxRadius = 0f

	fun updatedBorderPoints() {
		maxRadiusSquare = borderPoints.map { v -> v.lengthSquared() }.max() ?: 0f
		maxRadius = sqrt(maxRadiusSquare)
	}

	fun getCollisionPoly(): CollisionPolygon {
		return CollisionPolygon(this, borderPoints.map { v -> Vector2f(v).mulDirection(orientation).add(position) })
	}
}