package moonlander.views

import org.joml.Matrix4f
import org.lwjgl.opengl.GL30.*
import org.lwjgl.system.MemoryUtil
import moonlander.PerlinNoise


/**
 * OpenGL renderer for a single background "element" (e.g. a square from the background).
 */
@Suppress("EXPERIMENTAL_API_USAGE")
class BackgroundElementView(val indexX: Int, val indexY: Int, shouldInit: Boolean = true) : ImageObject(null) {

	companion object {
		const val elementSize = 30f
		val perlinNoiseGenerator = PerlinNoise()

		init {
			perlinNoiseGenerator.octaves.add(Pair(.5f, 1 / elementSize))
			perlinNoiseGenerator.octaves.add(Pair(.25f, 2 / elementSize))
			perlinNoiseGenerator.octaves.add(Pair(.125f, 4 / elementSize))
			perlinNoiseGenerator.octaves.add(Pair(.06f, 8 / elementSize))
		}
	}

	val w = 64 // resolution can be low, since perlin noise is blurred anyways.
	val h = w
	@ExperimentalUnsignedTypes
	var image: UByteArray = UByteArray(w * h * 4)

	init {
		println("Generating background tile for ($indexX, $indexY)")
		loadTexture("")
		if (shouldInit)
			init()
	}

	/**
	 * This initialization part must be executed in OpenGL's main thread.
	 */
	fun init() {
		createTexture()
		genTextureGeometryProgram(ar)
	}

	@ExperimentalUnsignedTypes
	override fun loadTexture(file: String) {
		for (x in 0 until w) {
			// interpolate height at this position (unit: m)
			val ex = (indexX + (x + .5f) / w) * elementSize

			for (y in 0 until h) {
				val ey = (indexY + (h - y - .5f) / h) * elementSize
				/* val scale = textureNoiseGenerator.eval(ex / 30.0, ey / 30.0) +
						0.5 * textureNoiseGenerator.eval(ex / 15.0, 1000 + ey / 15.0) +
						0.3 * textureNoiseGenerator.eval(ex / 7.5, 2000 + ey / 7.5) +
						0.15 * textureNoiseGenerator.eval(ex / 3.0, 3000 + ey / 3.0)*/
				val scale = perlinNoiseGenerator.getValue2d(ex, ey)
				image[(x + w * y) * 4] = (30 + 30 * scale).toByte().toUByte()
				image[(x + w * y) * 4 + 1] = 0u
				image[(x + w * y) * 4 + 2] = (70 + 70 * scale).toByte().toUByte()
				image[(x + w * y) * 4 + 3] = 255u
			}
		}

		ar = 1f
	}

	@ExperimentalUnsignedTypes
	fun createTexture() {
		// Init OpenGL texture
		tex = glGenTextures()
		glActiveTexture(tex)
		glBindTexture(GL_TEXTURE_2D, tex)
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE)
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE)
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR)
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR_MIPMAP_LINEAR)

		glEnable(GL_BLEND)
		glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA)

		val imageBuffer = MemoryUtil.memAlloc(image.size)
		imageBuffer.put(image.toByteArray())
		imageBuffer.position(0)

		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, w, h, 0, GL_RGBA, GL_UNSIGNED_BYTE, imageBuffer)

		glGenerateMipmap(GL_TEXTURE_2D)

		MemoryUtil.memFree(imageBuffer)
	}

	override fun render(viewMatrix: Matrix4f) {
		if (tex < 0)
			init()
		modelMatrix.identity()
			.translate((indexX + .5f) * elementSize, (indexY + .5f) * elementSize, 0f)
			.scale(elementSize / 2f, elementSize / 2f, 1f)
		super.render(viewMatrix)
	}

}