package moonlander.views

import org.joml.Matrix4f
import org.joml.Vector4f
import java.util.concurrent.Executors
import java.util.concurrent.LinkedBlockingQueue
import java.util.concurrent.ThreadPoolExecutor
import java.util.concurrent.TimeUnit
import kotlin.math.ceil
import kotlin.math.floor
import java.util.concurrent.ExecutorService


/**
 * Renders the background dust.
 * Background is split into square "elements" that can be generated and rendered individually (as the screen scrolls).
 * Generation might take some time, therefore it is parallized.
 */
class BackgroundView {

	private val backgroundElements = HashMap<Pair<Int, Int>, BackgroundElementView>()
	private val scheduledBackgroundElements = HashSet<Pair<Int, Int>>()

	fun render(viewMatrix: Matrix4f) {
		// Detect left / right border of viewport
		val inverseViewMatrix = Matrix4f()
		viewMatrix.invert(inverseViewMatrix)
		val unitVector = Vector4f(-1.05f, -1.05f, 0f, 1f)
		unitVector.mul(inverseViewMatrix)
		val leftBorder = unitVector.x
		val bottomBorder = unitVector.y
		unitVector.set(1.05f, 1.05f, 0f, 1f)
		unitVector.mul(inverseViewMatrix)
		val rightBorder = unitVector.x
		val topBorder = unitVector.y

		// ensure all ground elements are loaded, then render
		val x1 = floor(leftBorder / BackgroundElementView.elementSize).toInt()
		val x2 = ceil(rightBorder / BackgroundElementView.elementSize).toInt()
		val y1 = floor(bottomBorder / BackgroundElementView.elementSize).toInt()
		val y2 = ceil(topBorder / BackgroundElementView.elementSize).toInt()
		synchronized(backgroundElements) {
			for (x in x1..x2) {
				for (y in y1..y2) {
					val element = backgroundElements.getOrPut(Pair(x, y)) { BackgroundElementView(x, y) }
					element.render(viewMatrix)
				}
			}
		}
		preload(x1 - 1, x2 + 1, y1, y2)
	}


	/**
	 * Pre-generate some elements before the game is fully loaded
	 */
	fun preloadStart() {
		// val t = System.nanoTime()
		val x1 = -2
		val x2 = 2
		val y1 = -1
		val y2 = 2

		val executor = Executors.newFixedThreadPool(Runtime.getRuntime().availableProcessors())

		for (x in x1..x2) {
			for (y in y1..y2) {
				executor.submit {
					val e = BackgroundElementView(x, y, false)
					synchronized(backgroundElements) {
						backgroundElements.put(Pair(x, y), e)
					}
				}
			}
		}

		executor.shutdown()
		executor.awaitTermination(1000L, TimeUnit.SECONDS)
		for (e in backgroundElements.values)
			e.init()
		// println((System.nanoTime() - t) / 1000000000.0)
	}


	// private val backgroundExecutor = Executors.newFixedThreadPool(Runtime.getRuntime().availableProcessors()-1)
	private val backgroundExecutor =
		createFixedTimeoutExecutorService(Runtime.getRuntime().availableProcessors() - 1, 2, TimeUnit.SECONDS)

	fun preload(x1: Int, x2: Int, y1: Int, y2: Int) {
		synchronized(backgroundElements) {
			for (x in x1..x2) {
				for (y in y1..y2) {
					if (!backgroundElements.containsKey(Pair(x, y)) && !scheduledBackgroundElements.contains(
							Pair(
								x,
								y
							)
						)
					) {
						scheduledBackgroundElements.add(Pair(x, y))
						backgroundExecutor.submit {
							val e = BackgroundElementView(x, y, false)
							synchronized(backgroundElements) {
								backgroundElements.putIfAbsent(Pair(x, y), e)
							}
						}
					}
				}
			}
		}
	}


	/**
	 * Creates an executor service with a fixed pool size, that will time
	 * out after a certain period of inactivity.
	 *
	 * @param poolSize The core- and maximum pool size
	 * @param keepAliveTime The keep alive time
	 * @param timeUnit The time unit
	 * @return The executor service
	 */
	private fun createFixedTimeoutExecutorService(
		poolSize: Int,
		keepAliveTime: Long,
		timeUnit: TimeUnit
	): ExecutorService {
		val e = ThreadPoolExecutor(
			poolSize, poolSize,
			keepAliveTime, timeUnit, LinkedBlockingQueue()
		)
		e.allowCoreThreadTimeOut(true)
		return e
	}
}