package moonlander.views

import moonlander.models.Ground
import org.joml.Matrix4f
import org.lwjgl.opengl.GL30.*
import org.lwjgl.system.MemoryUtil
import kotlin.math.roundToInt


/**
 * Render a single "bucket" of the ground.
 */
@Suppress("EXPERIMENTAL_API_USAGE")
class GroundElementView(val ground: Ground, val index: Int) : ImageObject(null) {

	var zeroPosition: Float = 0f

	init {
		loadTexture("")
		genTextureGeometryProgram(ar)
	}

	@ExperimentalUnsignedTypes
	override fun loadTexture(file: String) {
		// Init OpenGL texture
		tex = glGenTextures()
		glActiveTexture(tex)
		glBindTexture(GL_TEXTURE_2D, tex)
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE)
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE)
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR)
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR_MIPMAP_LINEAR)

		glEnable(GL_BLEND)
		glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA)

		val w = 512
		val h = (w * ground.maxDiffHeight * 2f / ground.bucketWidth).toInt() // mountains
		val h2 = h // ground cover (always ground)
		// ((h+h2)/2 - h/2) / ((h+h2)/2)
		// => h2 / (h+h2)
		zeroPosition = h2.toFloat() / (h + h2)

		val image = UByteArray(w * (h + h2) * 4)
		for (x in 0 until w) {
			// interpolate height at this position
			val bucketPos = (x + .5f) * ground.bucketSize / w
			val heightOverLevel = ground.heightAtIndex(index, bucketPos)
			val max_y = ((1f - heightOverLevel / ground.maxDiffHeight) / 2f * h).roundToInt()
			val ex = (index + (x + .5f) / w) * ground.bucketWidth

			for (y in 0 until h + h2) {
				if (y <= max_y) {
					image[(x + w * y) * 4] = 0u
					image[(x + w * y) * 4 + 1] = 0u
					image[(x + w * y) * 4 + 2] = 0u
					image[(x + w * y) * 4 + 3] = if (y == max_y) 255u else 0u
				} else {
					val ey = ((h - y - .5f) / w) * ground.bucketWidth
					val scale = ground.perlinNoiseGenerator.getValue2d(ex, ey)
					val color = (100 + 90 * scale).toByte().toUByte()
					image[(x + w * y) * 4] = color
					image[(x + w * y) * 4 + 1] = color
					image[(x + w * y) * 4 + 2] = color
					image[(x + w * y) * 4 + 3] = 255u
				}
			}
		}

		val imageBuffer = MemoryUtil.memAlloc(image.size)
		imageBuffer.put(image.toByteArray())
		imageBuffer.position(0)

		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, w, h + h2, 0, GL_RGBA, GL_UNSIGNED_BYTE, imageBuffer)

		glGenerateMipmap(GL_TEXTURE_2D)

		MemoryUtil.memFree(imageBuffer)

		ar = 1f
	}

	override fun render(viewMatrix: Matrix4f) {
		modelMatrix.identity()
			.translate((index + .5f) * ground.bucketWidth, ground.level, 0f)
			.scale(ground.bucketWidth / 2f, ground.maxDiffHeight / (1f - zeroPosition), 1f)
			.translate(0f, -zeroPosition, 0f)
		super.render(viewMatrix)
	}
}