package moonlander.views

import org.joml.Matrix4f
import org.lwjgl.BufferUtils
import org.lwjgl.opengl.GL30.*
import org.lwjgl.stb.STBImage
import org.lwjgl.system.MemoryUtil

/**
 * OpenGL wrapper for:
 * Load an image containing multiple sub-images and render always one.
 *
 * The image will be split in rows and columns. Before each rendering you have to select the cell which is then going to be rendered.
 *
 * Can be used for cheap animations (showing a different part of the texture every frame).
 */
open class SpriteRenderer(resourceFile: String, val rows: Int, val cols: Int) {

	protected var ar = 0f
	private var vao: Int = 0
	private var vbo: Int = 0
	protected var tex: Int = 0
	private var program: Int = 0
	private var vertices: FloatArray = FloatArray(0)
	private var h0 = 0f
	private var h1 = 0f
	private var w0 = 0f
	private var w1 = 0f

	private var viewMatrixLocation: Int = 0
	private val viewMatrixBuffer = BufferUtils.createFloatBuffer(16)

	private var modelMatrixLocation: Int = 0
	private val modelMatrixBuffer = BufferUtils.createFloatBuffer(16)
	val modelMatrix = Matrix4f()

	init {
		loadTexture(resourceFile)
		genTextureGeometryProgram(ar)
	}

	private var lastRow = -1
	private var lastCol = -1

	fun submitBuffer(row: Int, col: Int) {
		if (row != lastRow || col != lastCol) {
			val u0 = col * 1f / cols
			val u1 = (col + 1) * 1f / cols
			val v0 = row * 1f / rows
			val v1 = (row + 1) * 1f / rows
			vertices[0 * 4 + 2] = u0
			vertices[0 * 4 + 3] = v1
			vertices[1 * 4 + 2] = u1
			vertices[1 * 4 + 3] = v0
			vertices[2 * 4 + 2] = u0
			vertices[2 * 4 + 3] = v0
			vertices[3 * 4 + 2] = u0
			vertices[3 * 4 + 3] = v1
			vertices[4 * 4 + 2] = u1
			vertices[4 * 4 + 3] = v1
			vertices[5 * 4 + 2] = u1
			vertices[5 * 4 + 3] = v0

			glBindBuffer(GL_ARRAY_BUFFER, vbo)
			glBufferSubData(GL_ARRAY_BUFFER, 0, vertices)

			lastRow = row
			lastCol = col
		}
	}

	open fun render(viewMatrix: Matrix4f) {
		glUseProgram(program)
		glBindVertexArray(vao)
		glBindTexture(GL_TEXTURE_2D, tex)
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
		modelMatrix.get(modelMatrixBuffer)
		glUniformMatrix4fv(modelMatrixLocation, false, modelMatrixBuffer)
		viewMatrix.get(viewMatrixBuffer)
		glUniformMatrix4fv(viewMatrixLocation, false, viewMatrixBuffer)
		glDrawArrays(GL_TRIANGLES, 0, 6)
	}


	protected open fun loadTexture(file: String) {
		val bytes = this.javaClass.getResource(file).readBytes()
		val imageBuffer = MemoryUtil.memAlloc(bytes.size)
		imageBuffer.put(bytes)
		imageBuffer.position(0)

		val wBuf = intArrayOf(0)
		val hBuf = intArrayOf(0)
		val comp = intArrayOf(0)

		// Decode the image
		val image = STBImage.stbi_load_from_memory(imageBuffer, wBuf, hBuf, comp, 0)
			?: throw RuntimeException("Failed to load image: " + STBImage.stbi_failure_reason()!!)
		val w = wBuf[0]
		val h = hBuf[0]

		// Init OpenGL texture
		tex = glGenTextures()
		glActiveTexture(tex)
		glBindTexture(GL_TEXTURE_2D, tex)
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE)
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE)
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR) // GL_LINEAR_MIPMAP_LINEAR - turn of filtering for now
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR) // GL_LINEAR_MIPMAP_LINEAR - turn of filtering for now

		glEnable(GL_BLEND)
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, w, h, 0, GL_RGBA, GL_UNSIGNED_BYTE, image)

		glGenerateMipmap(GL_TEXTURE_2D)

		MemoryUtil.memFree(imageBuffer)
		ar = w * 1f / h / cols * rows
	}

	protected fun genTextureGeometryProgram(ar: Float) {
		// Link between program and attributes, vbo, vertex array etc.
		vao = glGenVertexArrays()
		glBindVertexArray(vao)

		// Geometry - in range (-1, -1) - (1, 1)
		w1 = if (ar > 1) 1f else ar
		h1 = if (ar > 1) 1f / ar else 1f
		w0 = -w1
		h0 = -h1

		vertices = floatArrayOf(
			// x, y, u, v
			w0, h0, 0f, 1f,
			w1, h1, 1f, 0f,
			w0, h1, 0f, 0f,
			w0, h0, 0f, 1f,
			w1, h0, 1f, 1f,
			w1, h1, 1f, 0f
		)
		vbo = glGenBuffers()
		glBindBuffer(GL_ARRAY_BUFFER, vbo)
		glBufferData(GL_ARRAY_BUFFER, vertices, GL_DYNAMIC_DRAW)

		val shaderCode = """
			#version 150 core
			in vec2 position;
			in vec2 texcoord;

			uniform mat4 modelMatrix;
			uniform mat4 viewMatrix;

			out vec2 Texcoord;

			void main() {
				gl_Position = viewMatrix * modelMatrix * vec4(position, 0.0, 1.0);
				Texcoord = texcoord;
			}
		""".trimIndent()
		val vertexShader = glCreateShader(GL_VERTEX_SHADER)
		glShaderSource(vertexShader, shaderCode) // null?
		glCompileShader(vertexShader)
		val result = intArrayOf(0)
		glGetShaderiv(vertexShader, GL_COMPILE_STATUS, result)
		if (result[0] != GL_TRUE)
			println(glGetShaderInfoLog(vertexShader))

		val fragmentShaderCode = """
			#version 150 core
			in vec2 Texcoord;
			out vec4 outColor;

			uniform sampler2D tex;

			void main() {
				// outColor = vec4(1.0, 1.0, 0.5, 1.0);
				outColor = texture(tex, Texcoord); // * vec4(1.0, 1.0, 1.0, 1.0);
				//outColor[0] *= outColor[3];
				//outColor[1] *= outColor[3];
				//outColor[2] *= outColor[3];
			}
		""".trimIndent()
		val fragmentShader = glCreateShader(GL_FRAGMENT_SHADER)
		glShaderSource(fragmentShader, fragmentShaderCode) // null?
		glCompileShader(fragmentShader)
		glGetShaderiv(fragmentShader, GL_COMPILE_STATUS, result)
		if (result[0] != GL_TRUE)
			println(glGetShaderInfoLog(fragmentShader))

		// Program
		program = glCreateProgram()
		glAttachShader(program, vertexShader)
		glAttachShader(program, fragmentShader)
		glBindFragDataLocation(program, 0, "outColor")
		glLinkProgram(program)

		glUseProgram(program)

		// Making the link between vertex data and attributes
		// VBO is bound here!
		val posAttrib = glGetAttribLocation(program, "position")
		glVertexAttribPointer(posAttrib, 2, GL_FLOAT, false, 4 * 4, 0)
		glEnableVertexAttribArray(posAttrib)

		// Link texture attribute
		val texAttrib = glGetAttribLocation(program, "texcoord")
		glEnableVertexAttribArray(texAttrib)
		glVertexAttribPointer(texAttrib, 2, GL_FLOAT, false, 4 * 4, 2 * 4)

		modelMatrixLocation = glGetUniformLocation(program, "modelMatrix")
		viewMatrixLocation = glGetUniformLocation(program, "viewMatrix")
	}
}