package moonlander.views.hud

import moonlander.views.text.FontRenderer
import org.joml.Matrix4f
import org.lwjgl.BufferUtils
import org.lwjgl.opengl.GL30.*


/**
 * The health / fuel bars (including caption)
 */
class BarView(val width: Float = 1000f, val height: Float = 100f, val border: Float = 10f) {

	var text: String = ""
	var position: Float = 1f
	var colorBackground = floatArrayOf(.4f, 0f, 0f, 1f)
	var colorForeground = floatArrayOf(.8f, 0f, 0f, 1f)

	init {
		genGeometryProgram()
	}

	private var vao: Int = 0
	private var vbo: Int = 0
	private var program: Int = 0
	private var viewMatrixLocation: Int = 0
	private val viewMatrixBuffer = BufferUtils.createFloatBuffer(16)
	private var colorLocation: Int = 0

	protected fun genGeometryProgram() {
		// Link between program and attributes, vbo, vertex array etc.
		vao = glGenVertexArrays()
		glBindVertexArray(vao)
		vbo = glGenBuffers()
		glBindBuffer(GL_ARRAY_BUFFER, vbo)
		fillGeometryBuffer()

		val shaderCode = """
			#version 150 core
			in vec2 position;

			uniform mat4 viewMatrix;

			void main() {
				gl_Position = viewMatrix * vec4(position, 0.0, 1.0);
			}
		""".trimIndent()
		val vertexShader = glCreateShader(GL_VERTEX_SHADER)
		glShaderSource(vertexShader, shaderCode) // null?
		glCompileShader(vertexShader)
		val result = intArrayOf(0)
		glGetShaderiv(vertexShader, GL_COMPILE_STATUS, result)
		if (result[0] != GL_TRUE)
			println(glGetShaderInfoLog(vertexShader))

		val fragmentShaderCode = """
			#version 150 core
			out vec4 outColor;

			uniform vec4 color;

			void main() {
				outColor = color;
			}
		""".trimIndent()
		val fragmentShader = glCreateShader(GL_FRAGMENT_SHADER)
		glShaderSource(fragmentShader, fragmentShaderCode) // null?
		glCompileShader(fragmentShader)
		glGetShaderiv(fragmentShader, GL_COMPILE_STATUS, result)
		if (result[0] != GL_TRUE)
			println(glGetShaderInfoLog(fragmentShader))

		// Program
		program = glCreateProgram()
		glAttachShader(program, vertexShader)
		glAttachShader(program, fragmentShader)
		glBindFragDataLocation(program, 0, "outColor")
		glLinkProgram(program)

		glUseProgram(program)

		// Making the link between vertex data and attributes
		// VBO is bound here!
		val posAttrib = glGetAttribLocation(program, "position")
		glVertexAttribPointer(posAttrib, 2, GL_FLOAT, false, 2 * 4, 0)
		glEnableVertexAttribArray(posAttrib)

		colorLocation = glGetUniformLocation(program, "color")
		viewMatrixLocation = glGetUniformLocation(program, "viewMatrix")
	}


	fun fillGeometryBuffer() {
		val vertices = floatArrayOf(
			// x, y
			// outer (background)
			0f, 0f,
			width, height,
			0f, height,
			0f, 0f,
			width, 0f,
			width, height,
			// inner (bar)
			border, border,
			border + (width - 2 * border) * position, height - border,
			border, height - border,
			border, border,
			border + (width - 2 * border) * position, border,
			border + (width - 2 * border) * position, height - border
		)
		glBindBuffer(GL_ARRAY_BUFFER, vbo)
		glBufferData(GL_ARRAY_BUFFER, vertices, GL_STATIC_DRAW)
	}


	private val fontRenderer = FontRenderer(StatsHUDView.font, true, 30)

	fun render(viewMatrix: Matrix4f) {
		fillGeometryBuffer()

		glUseProgram(program)
		glBindVertexArray(vao)
		viewMatrix.get(viewMatrixBuffer)
		glUniformMatrix4fv(viewMatrixLocation, false, viewMatrixBuffer)
		glUniform4fv(colorLocation, colorBackground)
		glDrawArrays(GL_TRIANGLES, 0, 6)

		glUniform4fv(colorLocation, colorForeground)
		glDrawArrays(GL_TRIANGLES, 6, 6)

		fontRenderer.prepareTextBuffer(text, 5 * border, 1.5f * border)
		fontRenderer.render(viewMatrix)
	}

}