package moonlander.views.hud

import moonlander.Controller
import moonlander.format
import moonlander.views.text.FontRenderer
import org.joml.Matrix4f
import kotlin.math.floor
import kotlin.math.max

/**
 * Everything HUD-related (and how to render it).
 */
class HUDView(val controller: Controller) {

	// Top left text
	private val stats = StatsHUDView(controller)
	// Top right bars
	private val healthbar = BarView()
	private val fuelbar = BarView()

	// Intro / countdown
	private val beginText = BigHudMessage(10)
	private val beginText2 = BigHudMessage(160)
	// Success / error
	private val endscreenText = BigHudMessage(30)
	private val endscreenText2 = BigHudMessage(70)
	private val endscreenText3 = BigHudMessage(70)

	// Hit information top right below bars
	private val hitMessage1 = FontRenderer(StatsHUDView.font, true, 50)
	private val hitMessage2 = FontRenderer(StatsHUDView.font, true, 50)

	// green
	private val colorSuccess = floatArrayOf(0.4f, 1f, 0f, 1f)
	// orange
	private val colorError = floatArrayOf(1f, 0.4f, 0f, 1f)

	init {
		fuelbar.colorBackground = floatArrayOf(.2f, .4f, 0f, 1f)
		fuelbar.colorForeground = floatArrayOf(.35f, .7f, 0f, 1f)
		beginText.textcolor = colorSuccess
		endscreenText3.setText("< press [Enter] to restart >")
	}


	fun render(viewMatrix: Matrix4f, width: Float, height: Float) {
		stats.render(viewMatrix)

		// Render bars
		val position = viewMatrix.translate(width - 100f - healthbar.width, -100f - healthbar.height, 0f, Matrix4f())
		healthbar.text = "Health: ${controller.moonLander.hp}"
		healthbar.position = max(0.005f, controller.moonLander.hp / 100f)
		healthbar.render(position)

		position.translate(0f, -50f - fuelbar.height, 0f)
		fuelbar.text = "Fuel:   ${controller.moonLander.fuel.format(3, 1)} l"
		fuelbar.position = max(0.005f, controller.moonLander.fuel / controller.moonLander.maxFuel)
		fuelbar.render(position)


		// Render hit messages
		position.translate(healthbar.width, -100f, 0f)
		if ((controller.gameIsOver || controller.gameIsLost) && controller.moonLander.maxCollisionSpeedStands > 0f) {
			val text = "Maximal stand collision speed: ${controller.moonLander.maxCollisionSpeedStands.format(1, 1)} m/s"
			hitMessage1.prepareTextBuffer(text, -hitMessage1.getWidth(text).toFloat(), -hitMessage1.getHeight(text).toFloat())
			hitMessage1.textcolor =
				if (controller.moonLander.maxCollisionSpeedStands > controller.moonLander.standCollisionSpeedLimit) colorError else colorSuccess
			hitMessage1.render(position)
			position.translate(0f, -hitMessage1.getHeight(text) - 100f, 0f)
		}
		if ((controller.gameIsOver || controller.gameIsLost) && controller.moonLander.maxCollisionSpeedUpperLander > 0f) {
			val text = "Maximal payload collision speed: ${controller.moonLander.maxCollisionSpeedUpperLander.format(1, 1)} m/s"
			hitMessage2.prepareTextBuffer(text, -hitMessage2.getWidth(text).toFloat(), -hitMessage2.getHeight(text).toFloat())
			hitMessage2.textcolor =
				if (controller.moonLander.maxCollisionSpeedUpperLander > controller.moonLander.landerCollisionSpeedLimit) colorError else colorSuccess
			hitMessage2.render(position)
		}


		// Render initial screen
		if (controller.totalTime < 0) {
			if (controller.gameStarted) {
				val c = -floor(controller.totalTime).toInt()
				beginText.setText("<".repeat(c) + " $c " + ">".repeat(c))
				beginText2.setText("")
			} else {
				beginText.setText("< ready >")
				beginText2.setText("""
					[W], [space]  =  main engine (up)
					[A], [<-]     =  side engine (rotate left)
					[D], [->]     =  side engine (rotate right)

					      >>> press [space] to start <<<
				""".trimIndent())
			}
			viewMatrix.translate(width / 2, -height / 2, 0f, position)
			position.scale(2f, 2f, 2f)
			beginText.render(position)
			viewMatrix.translate(width / 2, -height / 2 - 500f, 0f, position)
			beginText2.render(position)
		}


		// Render endscreens
		if (controller.gameIsLost) {
			endscreenText.setText("<<< mission failed >>>")
			endscreenText.textcolor = colorError
			endscreenText2.setText("< ${controller.gameLostMessage} >")
			endscreenText2.textcolor = colorError
			endscreenText3.textcolor = colorError
		} else if (controller.gameIsOver) {
			endscreenText.setText("<<< mission accomplished >>>")
			endscreenText.textcolor = colorSuccess
			endscreenText2.setText("< lander ${controller.moonLander.hp}% operable >")
			endscreenText2.textcolor = colorSuccess
			endscreenText3.textcolor = colorSuccess
		}

		if (controller.gameIsOver || controller.gameIsLost) {
			viewMatrix.translate(width / 2, -height / 2, 0f, position)
			position.scale(2f, 2f, 2f)
			endscreenText.render(position)
			viewMatrix.translate(width / 2, -height / 2 - 250f, 0f, position)
			endscreenText2.render(position)
			viewMatrix.translate(width / 2, -height / 2 - 450f, 0f, position).scale(0.8f, 0.8f, 0.8f)
			endscreenText3.render(position)
		}
	}

}