package moonlander.views.text

// based on the example from https://github.com/SilverTiger/lwjgl3-tutorial/blob/master/src/silvertiger/tutorial/lwjgl/text/Font.java
// by Heiko Brumme

// Modified to fit in our context and render everything in one drawcall.

import moonlander.views.text.Texture.Companion.createTexture
import org.joml.Matrix4f
import org.lwjgl.BufferUtils
import org.lwjgl.opengl.GL30.*
import org.lwjgl.system.MemoryUtil
import java.awt.Color
import java.awt.Font
import java.awt.RenderingHints
import java.awt.geom.AffineTransform
import java.awt.image.AffineTransformOp
import java.awt.image.BufferedImage
import java.util.*


open class FontRenderer(font: Font, antiAlias: Boolean, max_chars: Int) {

	/**
	 * Contains the glyphs for each char.
	 */
	private val glyphs: MutableMap<Char, Glyph> = HashMap()
	/**
	 * Contains the font texture.
	 */
	private val texture: Texture

	/**
	 * Height of the font.
	 */
	private var fontHeight: Int = 0

	private val vertexBuffer: FloatArray = FloatArray(max_chars * 4 * 6)
	private var vertexBufferSize = 0
	private var vao = 0
	private var vbo = 0
	private var program = 0
	private var viewMatrixLocation: Int = 0
	private val viewMatrixBuffer = BufferUtils.createFloatBuffer(16)
	private var textcolorLocation: Int = 0

	var textcolor = floatArrayOf(1f, 1f, 1f, 1f)


	init {
		addVertexToBuffer(0f, 0f, 0f, 0f)
		addVertexToBuffer(1f, 0f, 0f, 0f)
		addVertexToBuffer(0f, 1f, 0f, 0f)
		texture = createFontTexture(font, antiAlias)
		genTextureGeometryProgram()
	}

	/**
	 * Creates a font texture from specified AWT font.
	 *
	 * @param font      The AWT font
	 * @param antiAlias Wheter the font should be antialiased or not
	 *
	 * @return Font texture
	 */
	private fun createFontTexture(font: Font, antiAlias: Boolean): Texture {
		/* Loop through the characters to get charWidth and charHeight */
		var imageWidth = 0
		var imageHeight = 0

		/* Start at char #32, because ASCII 0 to 31 are just control codes */
		for (i in 32..255) {
			if (i == 127) {
				/* ASCII 127 is the DEL control code, so we can skip it */
				continue
			}
			val c = i.toChar()
			val ch = createCharImage(font, c, antiAlias)
				?: /* If char image is null that font does not contain the char */
				continue

			imageWidth += ch.width
			imageHeight = Math.max(imageHeight, ch.height)
		}

		fontHeight = imageHeight

		/* Image for the texture */
		var image = BufferedImage(imageWidth, imageHeight, BufferedImage.TYPE_INT_ARGB)
		val g = image.createGraphics()

		var x = 0

		/* Create image for the standard chars, again we omit ASCII 0 to 31
         * because they are just control codes */
		for (i in 32..255) {
			if (i == 127) {
				/* ASCII 127 is the DEL control code, so we can skip it */
				continue
			}
			val c = i.toChar()
			val charImage = createCharImage(font, c, antiAlias)
				?: /* If char image is null that font does not contain the char */
				continue

			val charWidth = charImage.width
			val charHeight = charImage.height

			/* Create glyph and draw char on image */
			val ch = Glyph(charWidth, charHeight, x.toFloat(), (image.height - charHeight).toFloat())
			g.drawImage(charImage, x, 0, null)
			x += ch.width
			glyphs[c] = ch
		}

		/* Flip image Horizontal to get the origin to bottom left */
		val transform = AffineTransform.getScaleInstance(1.0, -1.0)
		transform.translate(0.0, (-image.height).toDouble())
		val operation = AffineTransformOp(
			transform,
			AffineTransformOp.TYPE_NEAREST_NEIGHBOR
		)
		image = operation.filter(image, null)

		/* Get charWidth and charHeight of image */
		val width = image.width
		val height = image.height

		/* Get pixel data of image */
		val pixels = IntArray(width * height)
		image.getRGB(0, 0, width, height, pixels, 0, width)

		/* Put pixel data into a ByteBuffer */
		val buffer = MemoryUtil.memAlloc(width * height * 4)
		for (i in 0 until height) {
			for (j in 0 until width) {
				/* Pixel as RGBA: 0xAARRGGBB */
				val pixel = pixels[i * width + j]
				/* Red component 0xAARRGGBB >> 16 = 0x0000AARR */
				buffer.put((pixel shr 16 and 0xFF).toByte())
				/* Green component 0xAARRGGBB >> 8 = 0x00AARRGG */
				buffer.put((pixel shr 8 and 0xFF).toByte())
				/* Blue component 0xAARRGGBB >> 0 = 0xAARRGGBB */
				buffer.put((pixel and 0xFF).toByte())
				/* Alpha component 0xAARRGGBB >> 24 = 0x000000AA */
				// buffer.put((pixel shr 24 and 0xFF).toByte())
				buffer.put((pixel and 0xFF).toByte())
			}
		}
		/* Do not forget to flip the buffer! */
		buffer.flip()

		/* Create texture */
		val fontTexture = createTexture(width, height, buffer)
		MemoryUtil.memFree(buffer)
		return fontTexture
	}

	/**
	 * Creates a char image from specified AWT font and char.
	 *
	 * @param font      The AWT font
	 * @param c         The char
	 * @param antiAlias Wheter the char should be antialiased or not
	 *
	 * @return Char image
	 */
	private fun createCharImage(font: Font, c: Char, antiAlias: Boolean): BufferedImage? {
		/* Creating temporary image to extract character size */
		var image = BufferedImage(1, 1, BufferedImage.TYPE_INT_ARGB)
		var g = image.createGraphics()
		if (antiAlias) {
			g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON)
		}
		g.font = font
		val metrics = g.fontMetrics
		g.dispose()

		/* Get char charWidth and charHeight */
		val charWidth = metrics.charWidth(c)
		val charHeight = metrics.height

		/* Check if charWidth is 0 */
		if (charWidth == 0) {
			return null
		}

		/* Create image for holding the char */
		image = BufferedImage(charWidth, charHeight, BufferedImage.TYPE_INT_ARGB)
		g = image.createGraphics()
		if (antiAlias) {
			g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON)
		}
		g.font = font
		g.paint = Color.WHITE
		g.drawString(c.toString(), 0, metrics.ascent)
		g.dispose()
		return image
	}

	/**
	 * Gets the width of the specified text.
	 *
	 * @param text The text
	 *
	 * @return Width of text
	 */
	fun getWidth(text: CharSequence): Int {
		var width = 0
		var lineWidth = 0
		for (i in 0 until text.length) {
			val c = text[i]
			if (c == '\n') {
				/* Line end, set width to maximum from line width and stored
                 * width */
				width = Math.max(width, lineWidth)
				lineWidth = 0
				continue
			}
			if (c == '\r') {
				/* Carriage return, just skip it */
				continue
			}
			val width1 = glyphs[c]?.width ?: 0
			lineWidth += width1
		}
		width = Math.max(width, lineWidth)
		return width
	}

	/**
	 * Gets the height of the specified text.
	 *
	 * @param text The text
	 *
	 * @return Height of text
	 */
	fun getHeight(text: CharSequence): Int {
		var height = 0
		var lineHeight = 0
		for (i in 0 until text.length) {
			val c = text[i]
			if (c == '\n') {
				/* Line end, add line height to stored height */
				height += lineHeight
				lineHeight = 0
				continue
			}
			if (c == '\r') {
				/* Carriage return, just skip it */
				continue
			}
			val height1 = glyphs[c]?.height ?: 0
			lineHeight = Math.max(lineHeight, height1)
		}
		height += lineHeight
		return height
	}

	/**
	 * Draw text at the specified position and color.
	 *
	 * @param text     Text to draw
	 * @param x        X coordinate of the text position
	 * @param y        Y coordinate of the text position
	 */
	fun prepareTextBuffer(text: CharSequence, x: Float, y: Float) {
		if (text.length*24 > vertexBuffer.size)
			throw RuntimeException("Text too long: ${text.length}")
		val textHeight = getHeight(text)

		var drawX = x
		var drawY = y
		if (textHeight > fontHeight) {
			drawY += textHeight - fontHeight
		}

		vertexBufferSize = 0

		texture.bind()
		for (i in 0 until text.length) {
			val ch = text[i]
			if (ch == '\n') {
				/* Line feed, set x and y to draw at the next line */
				drawY -= fontHeight
				drawX = x
				continue
			}
			if (ch == '\r') {
				continue
			}
			val (width, height, texX, texY) = glyphs[ch] ?: continue

			// render part of texture
			val x2 = drawX + width
			val y2 = drawY + height
			val s1 = texX / texture.width
			val t1 = texY / texture.height
			val s2 = (texX + width) / texture.width
			val t2 = (texY + height) / texture.height

			// put into vertex buffer
			addVertexToBuffer(drawX, drawY, s1, t1)
			addVertexToBuffer(drawX, y2, s1, t2)
			addVertexToBuffer(x2, y2, s2, t2)
			addVertexToBuffer(drawX, drawY, s1, t1)
			addVertexToBuffer(x2, y2, s2, t2)
			addVertexToBuffer(x2, drawY, s2, t1)

			drawX += width +1
		}
	}


	private fun addVertexToBuffer(x: Float, y: Float, s: Float, t: Float) {
		vertexBuffer[vertexBufferSize++] = x
		vertexBuffer[vertexBufferSize++] = y
		vertexBuffer[vertexBufferSize++] = s
		vertexBuffer[vertexBufferSize++] = t
	}

	private fun uploadVertexBuffer() {
		glBindBuffer(GL_ARRAY_BUFFER, vbo)
		glBufferSubData(GL_ARRAY_BUFFER, 0, vertexBuffer)
	}

	fun genTextureGeometryProgram() {
		vao = glGenVertexArrays()
		glBindVertexArray(vao)
		vbo = glGenBuffers()
		glBindBuffer(GL_ARRAY_BUFFER, vbo)
		glBufferData(GL_ARRAY_BUFFER, vertexBuffer, GL_DYNAMIC_DRAW)

		val shaderCode = """
			#version 150 core
			in vec2 position;
			in vec2 texcoord;

			uniform mat4 viewMatrix;

			out vec2 Texcoord;

			void main() {
				gl_Position = viewMatrix * vec4(position, 0.0, 1.0);
				// gl_Position = vec4(position, 0.0, 1.0);
				Texcoord = texcoord;
			}
		""".trimIndent()
		val vertexShader = glCreateShader(GL_VERTEX_SHADER)
		glShaderSource(vertexShader, shaderCode) // null?
		glCompileShader(vertexShader)
		val result = intArrayOf(0)
		glGetShaderiv(vertexShader, GL_COMPILE_STATUS, result)
		if (result[0] != GL_TRUE)
			println(glGetShaderInfoLog(vertexShader))

		val fragmentShaderCode = """
			#version 150 core
			in vec2 Texcoord;
			out vec4 outColor;

			uniform sampler2D tex;
			uniform vec4 textcolor;

			void main() {
				// outColor = vec4(1.0, 1.0, 0.5, 1.0);
				outColor = texture(tex, Texcoord) * textcolor;
			}
		""".trimIndent()
		val fragmentShader = glCreateShader(GL_FRAGMENT_SHADER)
		glShaderSource(fragmentShader, fragmentShaderCode) // null?
		glCompileShader(fragmentShader)
		glGetShaderiv(fragmentShader, GL_COMPILE_STATUS, result)
		if (result[0] != GL_TRUE)
			println(glGetShaderInfoLog(fragmentShader))

		// Program
		program = glCreateProgram()
		glAttachShader(program, vertexShader)
		glAttachShader(program, fragmentShader)
		glBindFragDataLocation(program, 0, "outColor")
		glLinkProgram(program)

		glUseProgram(program)

		// Making the link between vertex data and attributes
		// VBO is bound here!
		val posAttrib = glGetAttribLocation(program, "position")
		glVertexAttribPointer(posAttrib, 2, GL_FLOAT, false, 4 * 4, 0)
		glEnableVertexAttribArray(posAttrib)

		// Link texture attribute
		val texAttrib = glGetAttribLocation(program, "texcoord")
		glEnableVertexAttribArray(texAttrib)
		glVertexAttribPointer(texAttrib, 2, GL_FLOAT, false, 4 * 4, 2 * 4)

		viewMatrixLocation = glGetUniformLocation(program, "viewMatrix")
		textcolorLocation = glGetUniformLocation(program, "textcolor")
	}


	open fun render(viewMatrix: Matrix4f) {
		glUseProgram(program)
		glBindVertexArray(vao)
		texture.bind()
		glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA)
		uploadVertexBuffer()
		viewMatrix.get(viewMatrixBuffer)
		glUniformMatrix4fv(viewMatrixLocation, false, viewMatrixBuffer)
		glUniform4fv(textcolorLocation, textcolor)
		glDrawArrays(GL_TRIANGLES, 0, vertexBufferSize / 4)
	}

}