package moonlander.views.text

import java.nio.ByteBuffer
import org.lwjgl.system.MemoryStack

import org.lwjgl.opengl.GL30.*
import org.lwjgl.stb.STBImage.*
import org.lwjgl.opengl.EXTTextureFilterAnisotropic


/**
 * This class represents a texture.
 *
 * @author Heiko Brumme
 *
 * Used only for the FontRenderer, other textures are managed manually.
 */
class Texture {

	/**
	 * Stores the handle of the texture.
	 */
	private val id: Int = glGenTextures()

	/**
	 * Width of the texture.
	 */
	var width: Int = 0
		set(width) {
			if (width > 0) {
				field = width
			}
		}

	/**
	 * Height of the texture.
	 */
	var height: Int = 0
		set(height) {
			if (height > 0) {
				field = height
			}
		}

	/**
	 * Binds the texture.
	 */
	fun bind() {
		glBindTexture(GL_TEXTURE_2D, id)
	}

	/**
	 * Sets a parameter of the texture.
	 *
	 * @param name  Name of the parameter
	 * @param value Value to set
	 */
	fun setParameter(name: Int, value: Int) {
		glTexParameteri(GL_TEXTURE_2D, name, value)
	}

	/**
	 * Uploads image data with specified width and height.
	 *
	 * @param width  Width of the image
	 * @param height Height of the image
	 * @param data   Pixel data of the image
	 */
	fun uploadData(width: Int, height: Int, data: ByteBuffer) {
		uploadData(GL_RGBA8, width, height, GL_RGBA, data)
	}

	/**
	 * Uploads image data with specified internal format, width, height and
	 * image format.
	 *
	 * @param internalFormat Internal format of the image data
	 * @param width          Width of the image
	 * @param height         Height of the image
	 * @param format         Format of the image data
	 * @param data           Pixel data of the image
	 */
	fun uploadData(internalFormat: Int, width: Int, height: Int, format: Int, data: ByteBuffer) {
		glTexImage2D(GL_TEXTURE_2D, 0, internalFormat, width, height, 0, format, GL_UNSIGNED_BYTE, data)
	}

	/**
	 * Delete the texture.
	 */
	fun delete() {
		glDeleteTextures(id)
	}

	companion object {

		/**
		 * Creates a texture with specified width, height and data.
		 *
		 * @param width  Width of the texture
		 * @param height Height of the texture
		 * @param data   Picture Data in RGBA format
		 *
		 * @return Texture from the specified data
		 */
		fun createTexture(width: Int, height: Int, data: ByteBuffer): Texture {
			val texture = Texture()
			texture.width = width
			texture.height = height

			texture.bind()

			texture.setParameter(GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER)
			texture.setParameter(GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER)
			texture.setParameter(GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR)
			texture.setParameter(GL_TEXTURE_MAG_FILTER, GL_LINEAR_MIPMAP_LINEAR)
			val maxAnisotropy = glGetFloat(EXTTextureFilterAnisotropic.GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT)
			glTexParameterf(GL_TEXTURE_2D_ARRAY,EXTTextureFilterAnisotropic.GL_TEXTURE_MAX_ANISOTROPY_EXT,maxAnisotropy)

			texture.uploadData(GL_RGBA8, width, height, GL_RGBA, data)
			glGenerateMipmap(GL_TEXTURE_2D) // after uploading data!!!

			return texture
		}

		/**
		 * Load texture from file.
		 *
		 * @param path File path of the texture
		 *
		 * @return Texture from specified file
		 */
		fun loadTexture(path: String): Texture {
			var image: ByteBuffer? = null
			var width = 0
			var height = 0
			MemoryStack.stackPush().use { stack ->
				/* Prepare image buffers */
				val w = stack.mallocInt(1)
				val h = stack.mallocInt(1)
				val comp = stack.mallocInt(1)

				/* Load image */
				stbi_set_flip_vertically_on_load(true)
				image = stbi_load(path, w, h, comp, 4)
				if (image == null) {
					throw RuntimeException("Failed to load a texture file!" + System.lineSeparator() + stbi_failure_reason())
				}

				/* Get width and height of image */
				width = w.get()
				height = h.get()
			}

			return createTexture(width, height, image!!)
		}
	}

}
